# coding: utf-8
#
# SolidHarmonics.py is free software. It comes without any warranty, to the extent
# permitted by applicable law. You may use it, redistribute it and/or modify
# it, in whole or in part, provided that you do so at your own risk and do not
# hold the developers or copyright holders liable for any claim, damages, or
# other liabilities arising in connection with the software.
#
# This file is part of the IR integral core, and supplied as supporting
# information of the article:
#
#  [1] Mieke Peels and Gerald Knizia - "Fast evaluation of two-center integrals
#  over Gaussian charge distributions and Gaussian orbitals with general
#  interaction kernels"
#  J. Chem. Theory Comput. 2020, https://doi.org/10.1021/acs.jctc.9b01296
#
# - Gerald Knizia, 2019-12-29

"""This file contains a function to obtain symbolic expressions for the real
Solid Harmonics S_{l,m}(x,y,z) in the convention defined by 

   Helgaker, Jørgensen, Olsen - ``Molecular Electronic Structure Theory''
   (http://doi.org/10.1002/9781119019572)

in the Solid Harmonics chapter.

The concrete formulas implemented are:

  S_00 = 1     (recurrence starting point)

Diagonal recurrence relations:
  S_{l,l} = sqrt[ 2^(\delta_1l) (2l - 1)/(2l) ] *
          (x S_(l-1)(l-1) - (1-\delta_1l) y S_{(l-1),-(l-1)})

  S_{l,-l} = sqrt[ 2^(\delta_1l) (2l - 1)/(2l) ] *
          (y S_(l-1)(l-1) + (1-\delta_1l) x S_{l-1,-(l-1)})

Vertical recurrence relation:
  S_{l,m} = [(2l - 1) z S_(l-1)m - sqrt((l-1+m)(l-1-m)) r^2 S_{l-2,m}] *
          * 1/sqrt[(l+m)(l-m)].

r^2 here means x^2 + y^2 + z^2. S_{l,m} with l < 0 are considered zero.
"""
from __future__ import print_function
import sympy as sp
sy = sp.Symbol


_SlmMemo = {}
def Slm(l,m):
   """returns the cartesian expansion of real solid harmonic function S^l_m(x,y,z), as an analytic sympy expression."""
   if l < 0 or m < -l or m > l:
      return 0
   if (l,m) in _SlmMemo:
      return _SlmMemo[(l,m)]
   def delta(a,b):
      return 1 if a == b else 0
   def F(x):
      assert(type(x) is int or type(x) is str)
      return sp.sympify(x)
   
   Out = None
   x,y,z = sy('x'), sy('y'), sy('z')
   if l == 0:
      assert m == 0
      return F(1)
   elif l == m or l == -m:
      pf = sp.sqrt((F(2)**delta(1,l)*F(2*l - 1)/F(2*l)))
      if l == m:
         Out = pf * x * Slm(l-1,l-1) - pf * (1-delta(1,l)) * y * Slm(l-1,-(l-1))
      else:
         assert l == -m
         Out = pf * y * Slm(l-1,l-1) + pf * (1-delta(1,l)) * x * Slm(l-1,-(l-1))
   else:
      assert(abs(m) != l)
      pf = 1/sp.sqrt(F((l+m)*(l-m)))
      rsq = x**2 + y**2 + z**2
      #rsq = sy("rsq")
      Out = pf*(2*l - 1)*z*Slm(l-1,m) - pf*sp.sqrt(F((l-1+m)*(l-1-m))) * rsq * Slm(l-2,m)
   Out = sp.expand(Out)
   _SlmMemo[(l,m)] = Out
   return Out


def GetSlmCoeffs(l, m):
   """Express S^l_m(x,y,z) as a polynomial in x,y,z:
   
   S^l_m(x,y,z) = \sum_{ijk} s^{lm}_{ijk} x^i y^j z^k
   
   and return a dictionary mapping 3-tupples (i,j,k), representing the monomial
   exponents in x^i y^j z^k to the coefficients s^{lm}_{ijk}"""

   # get original polynomial with symbolc coefficients.
   x,y,z = sy('x'), sy('y'), sy('z')
   SlmX = sp.Poly(Slm(l,m), x, y, z)

   # collect prefactors into output dictionary
   NewTerms = {}
   for (Monomial, Coeff) in SlmX.terms():
      NewTerms[Monomial] = Coeff
   return NewTerms


def _SlmTest(MaxL):
   # if set, output numeric coefficients (e.g., 1.73205080756888) instead of analytic coefficients (e.g., 'sqrt(3)').
   #PrintNumericCoeffs = True
   PrintNumericCoeffs = False
   NumDigits = 16
   print(" ------------------\n -- _SlmTest()\n ------------------\n")
   print(" MAXL = {}  SYMBOLIC = {}  NUMDIGITS = {}".format(MaxL, not PrintNumericCoeffs, (NumDigits if PrintNumericCoeffs else "N/A")))
   def FormatCoeffs(c):
      if not PrintNumericCoeffs:
         # format as original symbolic sympy expression (do nothing)
         return c
      else:
         # numerically evaluate the constants up to the given precision
         return sp.N(c, NumDigits)
      pass
   
   for l in range(MaxL+1):
      print("\n Solid Harmonics for L = {}[{}]:".format(l, "spdfghiklmno"[l]))
      for m in range(-l,l+1):
         Slm_ = Slm(l,m)
         print("\n   S(l={},m={:+}) = {}".format(l,m, FormatCoeffs(Slm_)))
         for (Monomial, Coeff) in sorted(GetSlmCoeffs(l,m).items()):
            print("   |--> s^{{{},{}}}_{{{},{},{}}} = {}".format(l, m, Monomial[0], Monomial[1], Monomial[2], FormatCoeffs(Coeff)))
   pass


if __name__ == "__main__":
   _SlmTest(4)

